@@ -1411,6 +1411,7 @@ int amf_comp_response_1 (
 			assert (0);
 			break;
 	}
+	return 0;
 }
 
 /**
@@ -1687,4 +1688,6 @@ SaAmfReadinessStateT amf_comp_get_saAmfCompReadinessState (
 	}
 
 	assert (0);
+	/* XXX we fall here in case NDEBUG is set */
+	return -1;
 }
