@@ -66,6 +66,7 @@
 #include "jhash.h"
 #include "swab.h"
 #include "ipc.h"
+#include "flow.h"
 #include "print.h"
 
 #define GROUP_HASH_SIZE 32
@@ -76,7 +77,8 @@ enum cpg_message_req_types {
 	MESSAGE_REQ_EXEC_CPG_PROCJOIN = 0,
 	MESSAGE_REQ_EXEC_CPG_PROCLEAVE = 1,
 	MESSAGE_REQ_EXEC_CPG_JOINLIST = 2,
-	MESSAGE_REQ_EXEC_CPG_MCAST = 3
+	MESSAGE_REQ_EXEC_CPG_MCAST = 3,
+	MESSAGE_REQ_EXEC_CPG_DOWNLIST = 4
 };
 
 struct removed_group
@@ -102,6 +104,7 @@ struct process_info {
 	void *conn;
 	void *trackerconn;
 	struct group_info *group;
+	enum openais_flow_control_state flow_control_state;
 	struct list_head list; /* on the group_info members list */
 };
 
@@ -144,12 +147,18 @@ static void message_handler_req_exec_cpg_mcast (
 	void *message,
 	unsigned int nodeid);
 
+static void message_handler_req_exec_cpg_downlist (
+	void *message,
+	unsigned int nodeid);
+
 static void exec_cpg_procjoin_endian_convert (void *msg);
 
 static void exec_cpg_joinlist_endian_convert (void *msg);
 
 static void exec_cpg_mcast_endian_convert (void *msg);
 
+static void exec_cpg_downlist_endian_convert (void *msg);
+
 static void message_handler_req_lib_cpg_join (void *conn, void *message);
 
 static void message_handler_req_lib_cpg_leave (void *conn, void *message);
@@ -189,7 +198,7 @@ static struct openais_lib_handler cpg_lib_service[] =
 	},
 	{ /* 2 */
 		.lib_handler_fn				= message_handler_req_lib_cpg_mcast,
-		.response_size				= sizeof (mar_res_header_t),
+		.response_size				= sizeof (struct res_lib_cpg_mcast),
 		.response_id				= MESSAGE_RES_CPG_MCAST,
 		.flow_control				= OPENAIS_FLOW_CONTROL_REQUIRED
 	},
@@ -231,12 +240,17 @@ static struct openais_exec_handler cpg_exec_service[] =
 		.exec_handler_fn	= message_handler_req_exec_cpg_mcast,
 		.exec_endian_convert_fn	= exec_cpg_mcast_endian_convert
 	},
+	{ /* 4 */
+		.exec_handler_fn	= message_handler_req_exec_cpg_downlist,
+		.exec_endian_convert_fn	= exec_cpg_downlist_endian_convert
+	},
 };
 
 struct openais_service_handler cpg_service_handler = {
 	.name				        = (unsigned char*)"openais cluster closed process group service v1.01",
 	.id					= CPG_SERVICE,
 	.private_data_size			= sizeof (struct process_info),
+	.flow_control				= OPENAIS_FLOW_CONTROL_REQUIRED,
 	.lib_init_fn				= cpg_lib_init_fn,
 	.lib_exit_fn				= cpg_lib_exit_fn,
 	.lib_service				= cpg_lib_service,
@@ -304,9 +318,18 @@ struct req_exec_cpg_mcast {
 	mar_cpg_name_t group_name __attribute__((aligned(8)));
 	mar_uint32_t msglen __attribute__((aligned(8)));
 	mar_uint32_t pid __attribute__((aligned(8)));
+	mar_message_source_t source __attribute__((aligned(8)));
 	mar_uint8_t message[] __attribute__((aligned(8)));
 };
 
+struct req_exec_cpg_downlist {
+	mar_req_header_t header __attribute__((aligned(8)));
+	mar_uint32_t left_nodes __attribute__((aligned(8)));
+	mar_uint32_t nodeids[PROCESSOR_COUNT_MAX]  __attribute__((aligned(8)));
+};
+
+static struct req_exec_cpg_downlist req_exec_cpg_downlist;
+
 static void cpg_sync_init (void)
 {
 }
@@ -581,31 +604,53 @@ static void cpg_confchg_fn (
 	struct memb_ring_id *ring_id)
 {
 	int i;
-	struct list_head removed_list;
+	uint32_t lowest_nodeid = 0xffffff;
+	struct iovec req_exec_cpg_iovec;
 
-	log_printf(LOG_LEVEL_DEBUG, "confchg. joined_list: %d, left_list: %d\n", joined_list_entries, left_list_entries);
+	/* We don't send the library joinlist in here because it can end up
+	   out of order with the rest of the messages (which are totem ordered).
+	   So we get the lowest nodeid to send out a list of left nodes instead.
+	   On receipt of that message, all nodes will then notify their local clients
+	   of the new joinlist */
 
-	list_init(&removed_list);
+	if (left_list_entries) {
+		for (i = 0; i < member_list_entries; i++) {
+			if (member_list[i] < lowest_nodeid)
+				lowest_nodeid = member_list[i];
+		}
 
-	/* Remove nodes from joined groups and add removed groups to the list */
-	for (i = 0; i < left_list_entries; i++) {
-		remove_node_from_groups(left_list[i], &removed_list);
-	}
+		log_printf(LOG_LEVEL_DEBUG, "confchg, low nodeid=%d, us = %d\n", lowest_nodeid, this_ip->nodeid);
+		if (lowest_nodeid == this_ip->nodeid) {
 
-	if (!list_empty(&removed_list)) {
-		struct list_head *iter, *tmp;
+			req_exec_cpg_downlist.header.id = SERVICE_ID_MAKE(CPG_SERVICE, MESSAGE_REQ_EXEC_CPG_DOWNLIST);
+			req_exec_cpg_downlist.header.size = sizeof(struct req_exec_cpg_downlist);
 
-		for (iter = removed_list.next, tmp=iter->next; iter != &removed_list; iter = tmp, tmp = iter->next) {
-			struct removed_group *rg = list_entry(iter, struct removed_group, list);
-
-			notify_lib_joinlist(rg->gi, NULL,
-					    0, NULL,
-					    rg->left_list_entries, rg->left_list,
-					    MESSAGE_RES_CPG_CONFCHG_CALLBACK);
-			rg->gi->rg = NULL;
-			free(rg);
+			req_exec_cpg_downlist.left_nodes = left_list_entries;
+			for (i = 0; i < left_list_entries; i++) {
+				req_exec_cpg_downlist.nodeids[i] = left_list[i];
+			}
+			log_printf(LOG_LEVEL_DEBUG, "confchg, build downlist: %d nodes\n", left_list_entries);
 		}
 	}
+
+	/* Don't send this message until we get the final configuration message */
+	if (configuration_type == TOTEM_CONFIGURATION_REGULAR && req_exec_cpg_downlist.left_nodes) {
+		req_exec_cpg_iovec.iov_base = &req_exec_cpg_downlist;
+		req_exec_cpg_iovec.iov_len = req_exec_cpg_downlist.header.size;
+
+		totempg_groups_mcast_joined (openais_group_handle, &req_exec_cpg_iovec, 1, TOTEMPG_AGREED);
+		req_exec_cpg_downlist.left_nodes = 0;
+		log_printf(LOG_LEVEL_DEBUG, "confchg, sent downlist\n");
+	}
+}
+
+static void cpg_flow_control_state_set_fn (
+	void *context,
+	enum openais_flow_control_state flow_control_state)
+{
+	struct process_info *process_info = (struct process_info *)context;
+
+	process_info->flow_control_state = flow_control_state;
 }
 
 /* Can byteswap join & leave messages */
@@ -632,6 +677,13 @@ static void exec_cpg_joinlist_endian_convert (void *msg)
 	}
 }
 
+static void exec_cpg_downlist_endian_convert (void *msg)
+{
+	struct req_exec_cpg_downlist *req_exec_cpg_downlist = (struct req_exec_cpg_downlist *)msg;
+
+	req_exec_cpg_downlist->left_nodes = swab32(req_exec_cpg_downlist->left_nodes);
+}
+
 static void exec_cpg_mcast_endian_convert (void *msg)
 {
 	struct req_exec_cpg_mcast *req_exec_cpg_mcast = (struct req_exec_cpg_mcast *)msg;
@@ -640,7 +692,7 @@ static void exec_cpg_mcast_endian_convert (void *msg)
 	swab_mar_cpg_name_t (&req_exec_cpg_mcast->group_name);
 	req_exec_cpg_mcast->pid = swab32(req_exec_cpg_mcast->pid);
 	req_exec_cpg_mcast->msglen = swab32(req_exec_cpg_mcast->msglen);
-
+	swab_mar_message_source_t (&req_exec_cpg_mcast->source);
 }
 
 static void do_proc_join(
@@ -697,6 +749,39 @@ local_join:
 			    MESSAGE_RES_CPG_CONFCHG_CALLBACK);
 }
 
+static void message_handler_req_exec_cpg_downlist (
+	void *message,
+	unsigned int nodeid)
+{
+	struct req_exec_cpg_downlist *req_exec_cpg_downlist = (struct req_exec_cpg_downlist *)message;
+	int i;
+	struct list_head removed_list;
+
+	log_printf(LOG_LEVEL_DEBUG, "downlist left_list: %d\n", req_exec_cpg_downlist->left_nodes);
+
+	list_init(&removed_list);
+
+	/* Remove nodes from joined groups and add removed groups to the list */
+	for (i = 0; i <  req_exec_cpg_downlist->left_nodes; i++) {
+		remove_node_from_groups( req_exec_cpg_downlist->nodeids[i], &removed_list);
+	}
+
+	if (!list_empty(&removed_list)) {
+		struct list_head *iter, *tmp;
+
+		for (iter = removed_list.next, tmp=iter->next; iter != &removed_list; iter = tmp, tmp = iter->next) {
+			struct removed_group *rg = list_entry(iter, struct removed_group, list);
+
+			notify_lib_joinlist(rg->gi, NULL,
+					    0, NULL,
+					    rg->left_list_entries, rg->left_list,
+					    MESSAGE_RES_CPG_CONFCHG_CALLBACK);
+			rg->gi->rg = NULL;
+			free(rg);
+		}
+	}
+}
+
 static void message_handler_req_exec_cpg_procjoin (
 	void *message,
 	unsigned int nodeid)
@@ -782,11 +867,15 @@ static void message_handler_req_exec_cpg_mcast (
 {
 	struct req_exec_cpg_mcast *req_exec_cpg_mcast = (struct req_exec_cpg_mcast *)message;
 	struct res_lib_cpg_deliver_callback *res_lib_cpg_mcast;
+	struct process_info *process_info;
 	int msglen = req_exec_cpg_mcast->msglen;
 	char buf[sizeof(*res_lib_cpg_mcast) + msglen];
 	struct group_info *gi;
 	struct list_head *iter;
 
+	/*
+	 * Track local messages so that flow is controlled on the local node
+	 */
 	gi = get_group(&req_exec_cpg_mcast->group_name); /* this will always succeed ! */
 	assert(gi);
 
@@ -796,6 +885,12 @@ static void message_handler_req_exec_cpg_mcast (
 	res_lib_cpg_mcast->msglen = msglen;
 	res_lib_cpg_mcast->pid = req_exec_cpg_mcast->pid;
 	res_lib_cpg_mcast->nodeid = nodeid;
+	res_lib_cpg_mcast->flow_control_state = CPG_FLOW_CONTROL_DISABLED;
+	if (message_source_is_local (&req_exec_cpg_mcast->source)) {
+		openais_ipc_flow_control_local_decrement (req_exec_cpg_mcast->source.conn);
+		process_info = (struct process_info *)openais_conn_private_data_get (req_exec_cpg_mcast->source.conn);
+		res_lib_cpg_mcast->flow_control_state = process_info->flow_control_state;
+	}
 	memcpy(&res_lib_cpg_mcast->group_name, &gi->group_name,
 		sizeof(mar_cpg_name_t));
 	memcpy(&res_lib_cpg_mcast->message, (char*)message+sizeof(*req_exec_cpg_mcast),
@@ -911,6 +1006,14 @@ static void message_handler_req_lib_cpg_join (void *conn, void *message)
 		goto join_err;
 	}
 
+	openais_ipc_flow_control_create (
+		conn,
+		CPG_SERVICE,
+		req_lib_cpg_join->group_name.value,
+		req_lib_cpg_join->group_name.length,
+		cpg_flow_control_state_set_fn,
+		pi);
+
 	/* Add a node entry for us */
 	pi->nodeid = this_ip->nodeid;
 	pi->pid = req_lib_cpg_join->pid;
@@ -948,6 +1051,12 @@ static void message_handler_req_lib_cpg_leave (void *conn, void *message)
 	cpg_node_joinleave_send(gi, pi, MESSAGE_REQ_EXEC_CPG_PROCLEAVE, CONFCHG_CPG_REASON_LEAVE);
 	pi->group = NULL;
 
+	openais_ipc_flow_control_destroy (
+		conn,
+		CPG_SERVICE,
+		(unsigned char *)gi->group_name.value,
+		(unsigned int)gi->group_name.length);
+
 leave_ret:
 	/* send return */
 	res_lib_cpg_leave.header.size = sizeof(res_lib_cpg_leave);
@@ -964,7 +1073,7 @@ static void message_handler_req_lib_cpg_mcast (void *conn, void *message)
 	struct group_info *gi = pi->group;
 	struct iovec req_exec_cpg_iovec[2];
 	struct req_exec_cpg_mcast req_exec_cpg_mcast;
-	mar_res_header_t res;
+	struct res_lib_cpg_mcast res_lib_cpg_mcast;
 	int msglen = req_lib_cpg_mcast->msglen;
 	int result;
 
@@ -972,10 +1081,12 @@ static void message_handler_req_lib_cpg_mcast (void *conn, void *message)
 
 	/* Can't send if we're not joined */
 	if (!gi) {
-		res.size = sizeof(res);
-		res.id = MESSAGE_RES_CPG_MCAST;
-		res.error = SA_AIS_ERR_ACCESS; /* TODO Better error code ?? */
-		openais_conn_send_response(conn, &res, sizeof(res));
+		res_lib_cpg_mcast.header.size = sizeof(res_lib_cpg_mcast);
+		res_lib_cpg_mcast.header.id = MESSAGE_RES_CPG_MCAST;
+		res_lib_cpg_mcast.header.error = SA_AIS_ERR_ACCESS; /* TODO Better error code ?? */
+		res_lib_cpg_mcast.flow_control_state = CPG_FLOW_CONTROL_DISABLED;
+		openais_conn_send_response(conn, &res_lib_cpg_mcast,
+			sizeof(res_lib_cpg_mcast));
 		return;
 	}
 
@@ -984,6 +1095,7 @@ static void message_handler_req_lib_cpg_mcast (void *conn, void *message)
 		MESSAGE_REQ_EXEC_CPG_MCAST);
 	req_exec_cpg_mcast.pid = pi->pid;
 	req_exec_cpg_mcast.msglen = msglen;
+	message_source_set (&req_exec_cpg_mcast.source, conn);
 	memcpy(&req_exec_cpg_mcast.group_name, &gi->group_name,
 		sizeof(mar_cpg_name_t));
 
@@ -994,11 +1106,14 @@ static void message_handler_req_lib_cpg_mcast (void *conn, void *message)
 
 	// TODO: guarantee type...
 	result = totempg_groups_mcast_joined (openais_group_handle, req_exec_cpg_iovec, 2, TOTEMPG_AGREED);
+	openais_ipc_flow_control_local_increment (conn);
 
-	res.size = sizeof(res);
-	res.id = MESSAGE_RES_CPG_MCAST;
-	res.error = SA_AIS_OK;
-	openais_conn_send_response(conn, &res, sizeof(res));
+	res_lib_cpg_mcast.header.size = sizeof(res_lib_cpg_mcast);
+	res_lib_cpg_mcast.header.id = MESSAGE_RES_CPG_MCAST;
+	res_lib_cpg_mcast.header.error = SA_AIS_OK;
+	res_lib_cpg_mcast.flow_control_state = pi->flow_control_state;
+	openais_conn_send_response(conn, &res_lib_cpg_mcast,
+		sizeof(res_lib_cpg_mcast));
 }
 
 static void message_handler_req_lib_cpg_membership (void *conn, void *message)
