@@ -67,6 +67,7 @@
 #include "totemconfig.h"
 #include "main.h"
 #include "ipc.h"
+#include "flow.h"
 #include "service.h"
 #include "sync.h"
 #include "swab.h"
@@ -80,10 +81,25 @@
 
 #define SERVER_BACKLOG 5
 
+/*
+ * When there are this many entries left in a queue, turn on flow control
+ */
+#define FLOW_CONTROL_ENTRIES_ENABLE 400
+
+/*
+ * When there are this many entries in a queue, turn off flow control
+ */
+#define FLOW_CONTROL_ENTRIES_DISABLE 64
+
+
 static unsigned int g_gid_valid = 0;
 
 static struct totem_ip_address *my_ip;
 
+static totempg_groups_handle ipc_handle;
+
+DECLARE_LIST_INIT (conn_info_list_head);
+
 static void (*ipc_serialize_lock_fn) (void);
 
 static void (*ipc_serialize_unlock_fn) (void);
@@ -117,16 +133,22 @@ struct conn_info {
 	int authenticated;	/* Is this connection authenticated? */
 	void *private_data;	/* library connection private data */
 	struct conn_info *conn_info_partner;	/* partner connection dispatch<->response */
+	unsigned int flow_control_handle;	/* flow control identifier */
+	unsigned int flow_control_enabled;	/* flow control enabled bit */
+	unsigned int flow_control_local_count;	/* flow control local count */
+	enum openais_flow_control flow_control;	/* Does this service use IPC flow control */
+	pthread_mutex_t flow_control_mutex;
         int (*lib_exit_fn) (void *conn);
 	struct timerlist timerlist;
 	pthread_mutex_t mutex;
 	pthread_mutex_t *shared_mutex;
-
+	struct list_head list;
 };
 
 static void *prioritized_poll_thread (void *conn);
 static int conn_info_outq_flush (struct conn_info *conn_info);
 static void libais_deliver (struct conn_info *conn_info);
+static void ipc_flow_control (struct conn_info *conn_info);
 
  /*
   * IPC Initializers
@@ -155,7 +177,7 @@ static int response_init_send_response (
 	void *message)
 {
 	SaAisErrorT error = SA_AIS_ERR_ACCESS;
-	size_t cinfo = (size_t)conn_info;
+	uintptr_t cinfo = (uintptr_t)conn_info;
 	mar_req_lib_response_init_t *req_lib_response_init = (mar_req_lib_response_init_t *)message;
 	mar_res_lib_response_init_t res_lib_response_init;
 
@@ -185,7 +207,7 @@ static int dispatch_init_send_response (
 	void *message)
 {
 	SaAisErrorT error = SA_AIS_ERR_ACCESS;
-	size_t cinfo;
+	uintptr_t cinfo;
 	mar_req_lib_dispatch_init_t *req_lib_dispatch_init = (mar_req_lib_dispatch_init_t *)message;
 	mar_res_lib_dispatch_init_t res_lib_dispatch_init;
 	struct conn_info *msg_conn_info;
@@ -197,10 +219,19 @@ static int dispatch_init_send_response (
 		else
 			error = SA_AIS_OK;
 
-		cinfo = (size_t)req_lib_dispatch_init->conn_info;
+		cinfo = (uintptr_t)req_lib_dispatch_init->conn_info;
 		conn_info->conn_info_partner = (struct conn_info *)cinfo;
+
+		/* temporary fix for memory leak
+		 */
+		pthread_mutex_destroy (conn_info->conn_info_partner->shared_mutex);
+		free (conn_info->conn_info_partner->shared_mutex);
+		
 		conn_info->conn_info_partner->shared_mutex = conn_info->shared_mutex;
 
+		list_add (&conn_info_list_head, &conn_info->list);
+		list_add (&conn_info_list_head, &conn_info->conn_info_partner->list);
+
 		msg_conn_info = (struct conn_info *)cinfo;
 		msg_conn_info->conn_info_partner = conn_info;
 
@@ -245,6 +276,15 @@ static int dispatch_init_send_response (
 	conn_info->conn_info_partner->state = CONN_STATE_ACTIVE;
 	conn_info->lib_exit_fn = ais_service[conn_info->service]->lib_exit_fn;
 	ais_service[conn_info->service]->lib_init_fn (conn_info);
+
+	conn_info->flow_control = ais_service[conn_info->service]->flow_control;
+	conn_info->conn_info_partner->flow_control = ais_service[conn_info->service]->flow_control;
+	if (ais_service[conn_info->service]->flow_control == OPENAIS_FLOW_CONTROL_REQUIRED) {
+		openais_flow_control_ipc_init (
+			&conn_info->flow_control_handle,
+			conn_info->service);
+
+	}
 	return (0);
 }
 
@@ -283,15 +323,25 @@ static inline unsigned int conn_info_create (int fd) {
 	}
 
 	pthread_mutex_init (&conn_info->mutex, NULL);
+	pthread_mutex_init (&conn_info->flow_control_mutex, NULL);
 	pthread_mutex_init (conn_info->shared_mutex, NULL);
 
+	list_init (&conn_info->list);
 	conn_info->state = CONN_STATE_ACTIVE;
 	conn_info->fd = fd;
 	conn_info->events = POLLIN|POLLNVAL;
 	conn_info->service = SOCKET_SERVICE_INIT;
 
 	pthread_attr_init (&conn_info->thread_attr);
+/*
+ * IA64 needs more stack space then other arches
+ */
+#if defined(__ia64__)
+	pthread_attr_setstacksize (&conn_info->thread_attr, 400000);
+#else
 	pthread_attr_setstacksize (&conn_info->thread_attr, 200000);
+#endif
+
 	pthread_attr_setdetachstate (&conn_info->thread_attr, PTHREAD_CREATE_DETACHED);
 	res = pthread_create (&conn_info->thread, &conn_info->thread_attr,
 		prioritized_poll_thread, conn_info);
@@ -316,6 +366,11 @@ static void conn_info_destroy (struct conn_info *conn_info)
 	if (conn_info->conn_info_partner) {
 		conn_info->conn_info_partner->conn_info_partner = NULL;
 	}
+	
+	pthread_attr_destroy (&conn_info->thread_attr);
+	pthread_mutex_destroy (&conn_info->mutex);
+	
+	list_del (&conn_info->list);
 	free (conn_info);
 }
 
@@ -372,6 +427,9 @@ static int libais_disconnect (struct conn_info *conn_info)
 	}
 	conn_info->state = CONN_STATE_DISCONNECTED;
 	conn_info->conn_info_partner->state = CONN_STATE_DISCONNECTED;
+	if (conn_info->flow_control_enabled == 1) {
+		openais_flow_control_disable (conn_info->flow_control_handle);
+	}
 	return (0);
 }
 
@@ -410,9 +468,22 @@ static void *prioritized_poll_thread (void *conn)
 	pthread_mutex_t *rel_mutex;
 	unsigned int service;
 	struct conn_info *cinfo_partner;
+	void *private_data;
 
-	sched_param.sched_priority = 1;
-	res = pthread_setschedparam (conn_info->thread, SCHED_RR, &sched_param);
+#if defined(OPENAIS_BSD) || defined(OPENAIS_LINUX)
+	res = sched_get_priority_max (SCHED_RR);
+	if (res != -1) {
+		sched_param.sched_priority = res;
+		res = pthread_setschedparam (conn_info->thread, SCHED_RR, &sched_param);
+		if (res == -1) {
+			log_printf (LOG_LEVEL_WARNING, "Could not set SCHED_RR at priority %d: %s\n",
+				sched_param.sched_priority, strerror (errno));
+		}
+	} else
+		log_printf (LOG_LEVEL_WARNING, "Could not get maximum scheduler priority: %s\n", strerror (errno));
+#else
+	log_printf(LOG_LEVEL_WARNING, "Scheduler priority left to default value (no OS support)\n");
+#endif
 
 	ufd.fd = conn_info->fd;
 	for (;;) {
@@ -427,6 +498,7 @@ retry_poll:
 		switch (conn_info->state) {
 		case CONN_STATE_SECURITY:
 			conn_info_mutex_unlock (conn_info, service);
+			pthread_mutex_destroy (conn_info->shared_mutex);
 			free (conn_info->shared_mutex);
 			conn_info_destroy (conn);
 			pthread_exit (0);
@@ -443,6 +515,7 @@ retry_poll:
 
 		case CONN_STATE_DISCONNECTED:
 			rel_mutex = conn_info->shared_mutex;
+			private_data = conn_info->private_data;
 			cinfo_partner = conn_info->conn_info_partner;
 			conn_info_destroy (conn);
 			if (service == SOCKET_SERVICE_INIT) {
@@ -451,7 +524,9 @@ retry_poll:
 				pthread_mutex_unlock (rel_mutex);
 			}
 			if (cinfo_partner == NULL) {
+				pthread_mutex_destroy (rel_mutex);
 				free (rel_mutex);
+				free (private_data);
 			}
 			pthread_exit (0);
 			/*
@@ -487,6 +562,9 @@ retry_poll:
 			if ((ufd.revents & POLLIN) == POLLIN) {
 				libais_deliver (conn_info);
 			}
+
+			ipc_flow_control (conn_info);
+
 		}
 
 		ipc_serialize_unlock_fn ();
@@ -512,6 +590,64 @@ char *socketname = "/var/run/libais.socket";
 #endif
 
 
+static void ipc_flow_control (struct conn_info *conn_info)
+{
+	unsigned int entries_used;
+	unsigned int entries_usedhw;
+	unsigned int flow_control_local_count;
+	unsigned int fcc;
+
+	/*
+	 * Determine FCC variable and printing variables
+	 */
+	entries_used = queue_used (&conn_info->outq);
+	if (conn_info->conn_info_partner &&
+		queue_used (&conn_info->conn_info_partner->outq) > entries_used) {
+		entries_used = queue_used (&conn_info->conn_info_partner->outq);
+	}
+	entries_usedhw = queue_usedhw (&conn_info->outq);
+	if (conn_info->conn_info_partner &&
+		queue_usedhw (&conn_info->conn_info_partner->outq) > entries_used) {
+		entries_usedhw = queue_usedhw (&conn_info->conn_info_partner->outq);
+	}
+	flow_control_local_count = conn_info->flow_control_local_count;
+	if (conn_info->conn_info_partner &&
+		conn_info->conn_info_partner->flow_control_local_count > flow_control_local_count) {
+		flow_control_local_count = conn_info->conn_info_partner->flow_control_local_count;
+	}
+
+	fcc = entries_used;
+	if (flow_control_local_count > fcc) {
+		fcc = flow_control_local_count;
+	}
+	/*
+	 * IPC group-wide flow control
+	 */
+	if (conn_info->flow_control == OPENAIS_FLOW_CONTROL_REQUIRED) {
+		if (conn_info->flow_control_enabled == 0 &&
+			((fcc + FLOW_CONTROL_ENTRIES_ENABLE) > SIZEQUEUE)) {
+
+			log_printf (LOG_LEVEL_NOTICE, "Enabling flow control [%d/%d] - [%d].\n",
+				entries_usedhw, SIZEQUEUE,
+				flow_control_local_count);
+			openais_flow_control_enable (conn_info->flow_control_handle);
+			conn_info->flow_control_enabled = 1;
+			conn_info->conn_info_partner->flow_control_enabled = 1;
+		}
+		if (conn_info->flow_control_enabled == 1 &&
+
+			fcc <= FLOW_CONTROL_ENTRIES_DISABLE) {
+
+			log_printf (LOG_LEVEL_NOTICE, "Disabling flow control [%d/%d] - [%d].\n",
+				entries_usedhw, SIZEQUEUE,
+				flow_control_local_count);
+			openais_flow_control_disable (conn_info->flow_control_handle);
+			conn_info->flow_control_enabled = 0;
+			conn_info->conn_info_partner->flow_control_enabled = 0;
+		}
+	}
+}
+
 static int conn_info_outq_flush (struct conn_info *conn_info) {
 	struct queue *outq;
 	int res = 0;
@@ -574,6 +710,7 @@ retry_sendmsg:
 	if (queue_is_empty (outq)) {
 		conn_info->events = POLLIN|POLLNVAL;
 	}
+
 	return (0);
 }
 
@@ -632,7 +769,9 @@ static void libais_deliver (struct conn_info *conn_info)
 
 	iov_recv.iov_base = &conn_info->inb[conn_info->inb_start];
 	iov_recv.iov_len = (SIZEINB) - conn_info->inb_start;
-	assert (iov_recv.iov_len != 0);
+	if (conn_info->inb_inuse == SIZEINB) {
+		return;
+	}
 
 retry_recv:
 	res = recvmsg (conn_info->fd, &msg_recv, MSG_NOSIGNAL);
@@ -643,6 +782,12 @@ retry_recv:
 		return;
 	} else
 	if (res == 0) {
+#if defined(OPENAIS_SOLARIS) || defined(OPENAIS_BSD) || defined(OPENAIS_DARWIN)
+		/* On many OS poll never return POLLHUP or POLLERR.
+		 * EOF is detected when recvmsg return 0.
+		 */
+  		libais_disconnect_request (conn_info);
+#endif
 		return;
 	}
 
@@ -831,6 +976,15 @@ void message_source_set (
 	source->conn = conn;
 }
 
+static void ipc_confchg_fn (
+	enum totem_configuration_type configuration_type,
+	unsigned int *member_list, int member_list_entries,
+	unsigned int *left_list, int left_list_entries,
+	unsigned int *joined_list, int joined_list_entries,
+	struct memb_ring_id *ring_id)
+{
+}
+
 void openais_ipc_init (
 	void (*serialize_lock_fn) (void),
 	void (*serialize_unlock_fn) (void),
@@ -893,6 +1047,15 @@ void openais_ipc_init (
 	g_gid_valid = gid_valid;
 
 	my_ip = my_ip_in;
+
+	/*
+	 * Reset internal state of flow control when
+	 * configuration change occurs
+	 */
+	res = totempg_groups_initialize (
+		&ipc_handle,
+		NULL,
+		ipc_confchg_fn);
 }
 
 
@@ -947,6 +1110,9 @@ int openais_conn_send_response (
 	if (!libais_connection_active (conn_info)) {
 		return (-1);
 	}
+
+	ipc_flow_control (conn_info);
+
 	outq = &conn_info->outq;
 
 	msg_send.msg_iov = &iov_send;
@@ -1023,7 +1189,7 @@ retry_sendmsg_two:
 			conn_info->events = POLLIN|POLLNVAL;
 		}
 		if (res != -1) {
-			if (res + conn_info->byte_start != mlen) {
+			if (res != mlen) {
 				conn_info->byte_start += res;
 				res = -1;
 			} else {
@@ -1095,3 +1261,62 @@ void openais_ipc_timer_del_data (
 
 	timerlist_del (&conn_info->timerlist, timer_handle);
 }
+
+void openais_ipc_flow_control_create (
+	void *conn,
+	unsigned int service,
+	char *id,
+	int id_len,
+	void (*flow_control_state_set_fn) (void *conn, enum openais_flow_control_state),
+	void *context)
+{
+	struct conn_info *conn_info = (struct conn_info *)conn;
+
+	openais_flow_control_create (
+		conn_info->flow_control_handle,
+		service,
+		id,
+		id_len,
+		flow_control_state_set_fn,
+		context);	
+	conn_info->conn_info_partner->flow_control_handle = conn_info->flow_control_handle;
+}
+
+void openais_ipc_flow_control_destroy (
+	void *conn,
+	unsigned int service,
+	unsigned char *id,
+	int id_len)
+{
+	struct conn_info *conn_info = (struct conn_info *)conn;
+
+	openais_flow_control_destroy (
+		conn_info->flow_control_handle,
+		service,
+		id,
+		id_len);
+}
+
+void openais_ipc_flow_control_local_increment (
+        void *conn)
+{
+	struct conn_info *conn_info = (struct conn_info *)conn;
+
+	pthread_mutex_lock (&conn_info->flow_control_mutex);
+
+	conn_info->flow_control_local_count++;
+
+	pthread_mutex_unlock (&conn_info->flow_control_mutex);
+}
+
+void openais_ipc_flow_control_local_decrement (
+        void *conn)
+{
+	struct conn_info *conn_info = (struct conn_info *)conn;
+
+	pthread_mutex_lock (&conn_info->flow_control_mutex);
+
+	conn_info->flow_control_local_count--;
+
+	pthread_mutex_unlock (&conn_info->flow_control_mutex);
+}
