@@ -61,10 +61,10 @@
 #define TOKEN_TIMEOUT				1000
 #define TOKEN_RETRANSMIT_TIMEOUT		(int)(TOKEN_TIMEOUT / (TOKEN_RETRANSMITS_BEFORE_LOSS_CONST + 0.2))
 #define TOKEN_HOLD_TIMEOUT			(int)(TOKEN_RETRANSMIT_TIMEOUT * 0.8 - (1000/(int)HZ))
-#define JOIN_TIMEOUT				100
-#define CONSENSUS_TIMEOUT			200
+#define JOIN_TIMEOUT				50
+#define CONSENSUS_TIMEOUT			800
 #define MERGE_TIMEOUT				200
-#define DOWNCHECK_TIMEOUT			1000000
+#define DOWNCHECK_TIMEOUT			1000
 #define FAIL_TO_RECV_CONST			50
 #define	SEQNO_UNCHANGED_CONST			30
 #define MINIMUM_TIMEOUT				(int)(1000/HZ)*3
@@ -120,7 +120,6 @@ extern int totem_config_read (
 	char **error_string)
 {
 	int res = 0;
-	int parse_done = 0;
 	unsigned int object_totem_handle;
 	unsigned int object_interface_handle;
 	char *str;
@@ -129,7 +128,6 @@ extern int totem_config_read (
 	memset (totem_config, 0, sizeof (struct totem_config));
 	totem_config->interfaces = malloc (sizeof (struct totem_interface) * INTERFACE_MAX);
 	if (totem_config->interfaces == 0) {
-		parse_done = 1;
 		*error_string = "Out of memory trying to allocate ethernet interface storage area";
 		return -1;
 	}
