@@ -191,6 +191,8 @@ static void ckptSectionIterationInstanceFinalize (struct ckptSectionIterationIns
 
 	saHandleDestroy (&ckptSectionIterationHandleDatabase,
 		ckptSectionIterationInstance->sectionIterationHandle);
+
+	pthread_mutex_destroy (&ckptSectionIterationInstance->response_mutex);
 }
 
 static void ckptCheckpointInstanceFinalize (struct ckptCheckpointInstance *ckptCheckpointInstance)
@@ -214,6 +216,8 @@ static void ckptCheckpointInstanceFinalize (struct ckptCheckpointInstance *ckptC
 	list_del (&ckptCheckpointInstance->list);
 
 	saHandleDestroy (&checkpointHandleDatabase, ckptCheckpointInstance->checkpointHandle);
+
+	pthread_mutex_destroy (&ckptCheckpointInstance->response_mutex);
 }
 
 static void ckptInstanceFinalize (struct ckptInstance *ckptInstance)
@@ -235,6 +239,9 @@ static void ckptInstanceFinalize (struct ckptInstance *ckptInstance)
 	}
 
 	saHandleDestroy (&ckptHandleDatabase, ckptInstance->ckptHandle);
+
+	pthread_mutex_destroy (&ckptInstance->response_mutex);
+	pthread_mutex_destroy (&ckptInstance->dispatch_mutex);
 }
 
 /**
@@ -521,7 +528,7 @@ saCkptFinalize (
 	}
 
 	pthread_mutex_lock (&ckptInstance->response_mutex);
-
+	
 	/*
 	 * Another thread has already started finalizing
 	 */
@@ -534,7 +541,7 @@ saCkptFinalize (
 	ckptInstance->finalize = 1;
 
 	pthread_mutex_unlock (&ckptInstance->response_mutex);
-
+	
 	ckptInstanceFinalize (ckptInstance);
 
 	if (ckptInstance->response_fd != -1) {
@@ -1500,10 +1507,12 @@ saCkptCheckpointWrite (
 	for (i = 0; i < numberOfElements; i++) {
 		if (ioVector[i].dataSize == 0) {
 			*erroneousVectorIndex = i;
+			error = SA_AIS_ERR_INVALID_PARAM;
 			goto error_put;
 		}
 		if (ioVector[i].dataBuffer == NULL) {
 			*erroneousVectorIndex = i;
+			error = SA_AIS_ERR_INVALID_PARAM;
 			goto error_put;
 		}
 	}
@@ -1862,8 +1871,6 @@ error_put:
 	saHandleInstancePut (&checkpointHandleDatabase, checkpointHandle);
 
 	return (error == SA_AIS_OK ? res_lib_ckpt_checkpointsynchronizeasync.header.error : error);
-
-	return (SA_AIS_OK);
 }
 
 /** @} */
